'use strict'

/** @typedef {import('./index').Logger} Logger */

const chalk = require('chalk')
const Listr = require('listr')
const symbols = require('log-symbols')

const generateTasks = require('./generateTasks')
const getStagedFiles = require('./getStagedFiles')
const GitWorkflow = require('./gitWorkflow')
const makeCmdTasks = require('./makeCmdTasks')
const resolveGitDir = require('./resolveGitDir')

const debugLog = require('debug')('lint-staged:run')

/**
 * https://serverfault.com/questions/69430/what-is-the-maximum-length-of-a-command-line-in-mac-os-x
 * https://support.microsoft.com/en-us/help/830473/command-prompt-cmd-exe-command-line-string-limitation
 * https://unix.stackexchange.com/a/120652
 */
const MAX_ARG_LENGTH =
  (process.platform === 'darwin' && 262144) || (process.platform === 'win32' && 8191) || 131072

/**
 * Executes all tasks and either resolves or rejects the promise
 *
 * @param {object} options
 * @param {Object} [options.config] - Task configuration
 * @param {Object} [options.cwd] - Current working directory
 * @param {boolean} [options.relative] - Pass relative filepaths to tasks
 * @param {boolean} [options.shell] - Skip parsing of tasks for better shell support
 * @param {boolean} [options.quiet] - Disable lint-staged’s own console output
 * @param {boolean} [options.debug] - Enable debug mode
 * @param {Logger} logger
 * @returns {Promise}
 */
module.exports = async function runAll(
  { config, cwd = process.cwd(), debug = false, quiet = false, relative = false, shell = false },
  logger = console
) {
  debugLog('Running all linter scripts')

  const gitDir = await resolveGitDir({ cwd })

  if (!gitDir) {
    throw new Error('Current directory is not a git directory!')
  }

  debugLog('Resolved git directory to be `%s`', gitDir)

  const files = await getStagedFiles({ cwd: gitDir })

  if (!files) {
    throw new Error('Unable to get staged files!')
  }

  debugLog('Loaded list of staged files in git:\n%O', files)

  const argLength = files.join(' ').length
  if (argLength > MAX_ARG_LENGTH) {
    logger.warn(`
 ${symbols.warning}  ${chalk.yellow(
      `lint-staged generated an argument string of ${argLength} characters, and commands might not run correctly on your platform.
    It is recommended to use functions as linters and split your command based on the number of staged files. For more info, please visit:
    https://github.com/okonet/lint-staged#using-js-functions-to-customize-linter-commands`
    )}
`)
  }

  const tasks = generateTasks({ config, cwd, gitDir, files, relative })

  // lint-staged 10 will automatically add modifications to index
  // Warn user when their command includes `git add`
  let hasDeprecatedGitAdd = false

  const listrTasks = tasks.map(task => {
    const subTasks = makeCmdTasks({ commands: task.commands, files: task.fileList, gitDir, shell })

    if (subTasks.some(subTask => subTask.command.includes('git add'))) {
      hasDeprecatedGitAdd = true
    }

    return {
      title: `Running tasks for ${task.pattern}`,
      task: async () =>
        new Listr(subTasks, {
          // In sub-tasks we don't want to run concurrently
          // and we want to abort on errors
          dateFormat: false,
          concurrent: false,
          exitOnError: true
        }),
      skip: () => {
        if (task.fileList.length === 0) {
          return `No staged files match ${task.pattern}`
        }
        return false
      }
    }
  })

  if (hasDeprecatedGitAdd) {
    logger.warn(`${symbols.warning} ${chalk.yellow(
      `Some of your tasks use \`git add\` command. Please remove it from the config since all modifications made by tasks will be automatically added to the git commit index.`
    )}
`)
  }

  // If all of the configured tasks should be skipped
  // avoid executing any lint-staged logic
  if (listrTasks.every(task => task.skip())) {
    logger.log('No staged files match any of provided globs.')
    return 'No tasks to run.'
  }

  const listrOptions = {
    dateFormat: false,
    renderer: (quiet && 'silent') || (debug && 'verbose') || 'update'
  }

  const git = new GitWorkflow(gitDir)

  const runner = new Listr(
    [
      {
        title: 'Preparing...',
        task: () => git.stashBackup()
      },
      {
        title: 'Running tasks...',
        task: () => new Listr(listrTasks, { ...listrOptions, concurrent: true, exitOnError: false })
      },
      {
        title: 'Applying modifications...',
        skip: ctx => ctx.hasErrors && 'Skipped because of errors from tasks',
        task: () => git.applyModifications()
      },
      {
        title: 'Reverting to original state...',
        enabled: ctx => ctx.hasErrors,
        task: () => git.restoreOriginalState()
      },
      {
        title: 'Cleaning up...',
        task: () => git.dropBackup()
      }
    ],
    listrOptions
  )

  try {
    await runner.run()
  } catch (error) {
    if (error.message.includes('Another git process seems to be running in this repository')) {
      logger.error(`
  ${symbols.error} ${chalk.red(`lint-staged failed due to a git error.
    Any lost modifications can be restored from a git stash:

    > git stash list
    stash@{0}: On master: automatic lint-staged backup
    > git stash pop stash@{0}`)}
`)
    }

    throw error
  }
}
